// Name: main.c
// char    MYD_id[]=
// " sphNtest.C 2009/10/24  MORIYOSHI DENSHI CO.,LTD. COPYRIGHT(C) OCT.2009 ";
// License: GNU GPL v2
// avr ATmega168 12MHz
// fuse L:E0(11100000) H:DF(11011111) E:01(00000001) // External Clock
//   or L:C0(11100000) H:DD(11011111) E:01(00000001) // External Clock(BOD on)
// fuse L:FF(11111111) H:DF(11011111) E:01(00000001) // XTAL
//   or L:CF(11001111) H:DF(11011111) E:01(00000001) // CEROLOCK

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <string.h>

#ifndef uchar
#define uchar   unsigned char
#endif

#define WORD unsigned int

uchar consts, conibf[3];

void  coninit() {
    // initialize hardware
    PORTB = 0xB7;    // all pullups except PB3. PB6 external clock
    DDRB = (1<<PB4); // all inputs include PB3(~RTS) except PB4(~CTS) output
    PORTC = 0x3F;    // all pullups. PC6 cpu reset
    PORTD = 0xFE;    // all pullups
    DDRD = (1<<PD1); // all inputs include RXD(PD0 TxD) except TXD(PD1 RxD)

    // USART configuration : 9600bps 8N1
    UBRR0L  = 78-1;
    UCSR0B  = (1<<RXEN0) | (1<<TXEN0) | (1<<RXCIE0);
    consts = 0;
}

uchar conins() { // return val. : ready
    cli();
    if(!consts && !(UCSR0A & (1<<RXC0)) ) PORTB &= ~(1<<PB4);
    sei();
    return consts;
}

uchar coninp() { uchar dat;
    while( !conins()) ;
    cli();
    dat = conibf[0];
    conibf[0] = conibf[1];
    conibf[1] = conibf[2];
    --consts;
//@ if(consts == 0) PORTB &= ~(1<<PB4);
    UCSR0B |= (1<<RXCIE0);
    sei();
    return dat;
}

uchar conots() { // return val. : not ready
//  conintr();
    return (~UCSR0A & (1<<UDRE0)) | (PINB & (1<<PB3));
}

void  conout(uchar dat) {
    while( conots()) ;
    UDR0 = dat;
}

void  crlf() {
//  conout('\r'); // 0x0D enter
    conout('\n'); // 0x0A ^J
}

void  hxd(uchar dat) {
    dat &= 0x0F;
    if(dat >= 10) dat += ('A'-'0'-10);
    conout(dat + '0');
}

void  lbyte(uchar dat) {
    hxd(dat >> 4);
    hxd(dat);
}

void  ladr(WORD dat) {
    lbyte(dat >> 8);
    lbyte(dat & 0x00FF);
}

void  mesot(char *str) { uchar dat;
    while(1) {
        dat = pgm_read_byte(str);
        if(!dat) break;
        conout(dat); str++;
    }
}

uchar txtin() { uchar dat;
    dat = coninp();
    if(dat < 0x20) return dat; // control code
    conout(dat);
    if(dat >= 'a' && dat < 'z'+1) dat -= 'a'-'A';
    return dat;
}

uchar nibbl(uchar dat) {
    dat -= '0';
    if(dat < 10) return dat;
    dat += '0'-'A'+10;
    if(dat >= 10 && dat < 16) return dat;
    return 0x10;
}

extern uchar setdlmnbl(uchar *delmt, uchar ch);
/*uchar setdlmnbl(uchar *delmt, uchar ch) {
    *delmt = ch;
    return nibbl(ch);
}*/

WORD  expr1(uchar *delmt) { WORD dat; uchar ch;
    dat = 0;
    while(1) {
        ch = txtin();
//      *delmt = ch;
//      ch = nibbl(ch);
        ch = setdlmnbl(delmt, ch);
        if(ch >= 0x10) break;
        dat <<= 4;
        dat |= ch;
    }
    return dat;
}

PROGMEM char headmes[] = "avr HM-monitor\n";

int main(void) {
    uchar ch;

    coninit();
    sei();
ts: while(1) {
        ch = coninp();
        if(ch == 0x1B) break; // esc
        conout(ch);
        if(ch != '\n') conout(ch); // 0x0A
    }
    crlf();
    mesot((char *)headmes);
    while(1) {
        conout('.');
        ch = txtin();
        if(ch == 0x03) goto ts; // ^C
        if(ch == 'I') { WORD adr;
            adr = expr1(&ch);
            if(ch < 0x20) conout(' ');
            ladr(adr);
            conout('=');
            lbyte(_MMIO_BYTE(adr)); // 0x20-5F:IO 0x60-DF:RAM
            if(ch == '\r') crlf();
        }else{
            if(ch == '\r') crlf();
            if(ch == '\n') conout(' ');
            conout('x');
            lbyte(ch);
        }
    }
}

/* check register save(push,pop) in MtestASM.S by interrupt */
void rxrdyint0() {
    if(UCSR0A & (1<<RXC0)) {
        PORTB |= (1<<PB4);
        if(consts >= 3) consts = 2;
        conibf[consts++] = UDR0;
        if(consts >= 3) UCSR0B &= ~(1<<RXCIE0);
    }
}
